/*!
  @file rx_ssm.h
	
  @brief Screen Space Mesh쐬
 		- M. Muller et al., Screen space meshes, SCA2007, 2007. 

  @author Makoto Fujisawa
  @date 2011-05
*/
// FILE --rx_ssm.h--

#ifndef _RX_SSM_H_
#define _RX_SSM_H_

#pragma warning (disable: 4244)


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <cmath>
#include <cstdlib>

#include <iostream>
#include <fstream>

// STL
#include <vector>
#include <string>

// OpenGL
#include <GL/glut.h>

// f[^\
#include "rx_utility.h"
#include "rx_matrix.h"

// Marching Squaree[u
#include "rx_ssm_tables.h"

#include "rx_ssm_kernel.cuh"
#include <cutil_inline.h>
//#include <cutil_math.h>

using namespace std;



//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
#ifndef RXREAL
	#define RXREAL float
#endif
#ifndef RXREAL2
	#define RXREAL2 float2
#endif
#ifndef RXREAL3
	#define RXREAL3 float3
#endif




#define RX_USE_TIMER

#ifdef RX_USE_TIMER
	class rxTimerAvg;
	extern rxTimerAvg g_Time;

	#define RXTIMER_CLEAR g_Time.Clear()
	#define RXTIMER_RESET g_Time.ResetTime()
	#define RXTIMER(x) g_Time.Split(x); g_Time.ResetTime()
	#define RXTIMER_PRINT g_Time.Print()
	#define RXTIMER_STRING(x) g_Time.PrintToString(x)
#else
	#define RXTIMER_CLEAR
	#define RXTIMER_RESET
	#define RXTIMER(x) 
	#define RXTIMER_PRINT
	#define RXTIMER_STRING(x)
#endif

//! Screen Spaceł̃p[eBN
struct rxSSParticle
{
	Vec3 xp;	//!< XN[Xy[Xł̒SW
	Vec3 rp;	//!< XN[Xy[Xł̔a
};

//! Screen Space ł̒_
struct rxSSVertex
{
	Vec3 pos;		//!< Screen SpaceWl
	RXREAL depth;	//!< fvXl
	int edge;		//!< GbW_ł邩ǂ

	Vec3 avg_pos;	//!< אڒ_ύWl(֊sp)
	RXREAL avg_num;	//!< אڒ_(֊sp)

	rxSSVertex(Vec3 p, int e = 0) : pos(Vec3(p[0], p[1], p[2])), depth(p[2]), edge(e) {}
};

//! ֊sGbW
struct rxSSEdge
{
	Vec3 x0, x1;	//!< [_W
	RXREAL depth;	//!< GbWfvXl
	RXREAL d0, d1;	//!< [_fvXl
	int xy;			//!< GbW(x:0, y:1)
	int silhouette;	
	int front_vertex;	//!< GbW_̃CfbNX
	RXREAL dx;			//!< fvXl[_GbW_܂ł̋
};

//! bVpObh
struct rxSSGrid
{
	int i, j;
	int node_vrts[4];	//!< m[h_CfbNX
	int num_nv;			//!< m[h_
	int edge_vrts[4];	//!< GbW_(front vertex)
	int num_ev;			//!< GbW_(front vertex)
	int back_vrts[6];	//!< GbW_(back vertex, back-2 vertex)
	int num_bv;			//!< GbW_(back vertex)

	RXREAL node_depth[4];	//!< m[h̃fvXl

	int table_index0;	//!< fobOp:bV̂߂̃CfbNXl
	int table_index1;	//!< fobOp:bV̂߂̃CfbNXl
	int mesh_num;		//!< fobOp:bV
	int mesh[6];		//!< fobOp:bVCfbNX
	int back2;			//!< fobOp
	int v[14];			//!< fobOp
};


// m[h_
// 3 - 2
// |   |
// 0 - 1
	 
// GbW_
// - 2 -
// 3   1
// - 0 -


//-----------------------------------------------------------------------------
// ֐vg^Cv錾
//-----------------------------------------------------------------------------
inline RXREAL GetDepthNearestT(double x, double y, int nx, int ny, double dx, double dy, const vector<RXREAL> &dmap);
inline RXREAL GetDepthInterpT(double x, double y, int nx, int ny, double dx, double dy, const vector<RXREAL> &dmap);


//-----------------------------------------------------------------------------
// Screen Space Mesh - NX
// MARK:rxSSMesh
//-----------------------------------------------------------------------------
class rxSSMesh
{

protected:
	// oϐ
	double m_fDmx, m_fDmy;	//!< fvX}bv̊eObh
	int m_iNumNodeVrts;		//!< m[h_
	int m_iNumEdgeVrts;		//!< GbW_
	int m_iNumMesh;			//!< bV

	double m_fSpacing;		//!< fvX}bṽTvOԊu
	double m_fPrtRad;		//!< p[eBN̔a
	double m_fSSZmax;		//!< ֊sƂȂfvX臒l
	int m_iNfilter;			//!< fvXl̃tB^TCY
	int m_iNiters;			//!< ֊s̔
	int m_iNgx, m_iNgy;		//!< bVpObh̉𑜓x

protected:
	//! ftHgRXgN^
	rxSSMesh(){}	// wȂŃIuWFNg쐬̂h

public:
	//! RXgN^
	rxSSMesh(double zmax, double h, double r, int n_filter, int n_iter)
	{
		m_fSSZmax = zmax;
		m_fSpacing = h;
		m_fPrtRad = r;
		m_iNfilter = n_filter;
		m_iNiters = n_iter;
		m_iNgx = m_iNgy = 24;

		m_iNumNodeVrts = 0;
		m_iNumEdgeVrts = 0;
		m_iNumMesh = 0;
	}

	//! fXgN^
	virtual ~rxSSMesh()
	{
	}

public:
	//
	// ANZX\bh
	//
	//! Obh
	int GetNx(void) const { return m_iNgx; }
	int GetNy(void) const { return m_iNgy; }
	void GetDepthMapSize(int &nx, int &ny){ nx = m_iNgx+1; ny = m_iNgy+1; }

	//! fvX}bṽTvOԊu
	void   SetSpacing(double h){ m_fSpacing = h; }
	double GetSpacing(void) const { return m_fSpacing; }

	//! p[eBNa
	void   SetRadius(double r){ m_fPrtRad = r; }
	double GetRadius(void) const { return m_fPrtRad; }

	//! ֊sƂȂfvX臒l
	void   SetZMax(double r){ m_fSSZmax = r; }
	double GetZMax(void) const { return m_fSSZmax; }

	//! fvXl̃tB^TCY
	void SetFilterRadius(int r){ m_iNfilter = r; }
	int  GetFilterRadius(void) const { return m_iNfilter; }

	//! ֊s̔
	void SetSmoothIter(int r){ m_iNiters = r; }
	int  GetSmoothIter(void) const { return m_iNiters; }

	//! fvX}bv
	virtual RXREAL* GetDepthMap(void) = 0;

	//! _
	virtual rxSSVertex* GetSSVertex(void) = 0;

	//! bVpObh
	virtual rxSSGrid GetMeshGrid(int idx) = 0;
	virtual rxSSGrid GetMeshGrid(int i, int j) = 0;

public:
	//
	// ֐
	//

	/*!
	 * XN[Xy[XbV
	 * @param[in] proj OpenGLes
	 * @param[in] modelview OpenGLfr[s
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] prts p[eBNW
	 * @param[in] pnum p[eBN
	 * @param[out] vrts bV_
	 * @param[out] polys bV
	 * @param[in] filtering tB^OtO(fvXl0x01, ֊s0x02)
	 * @param[in] debug_output ʂ̉ʏo̗͂L
	 */
	virtual void CreateMesh(double *proj, double *modelview, int W, int H, vector<Vec3> &prts, int pnum, 
							vector<Vec3> &vrts, vector< vector<int> > &mesh, int filtering = 0, int debug_output = 0) = 0;

	/*!
	 * XN[Xy[XbV(@vZ܂)
	 * @param[in] proj OpenGLes
	 * @param[in] modelview OpenGLfr[s
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] prts p[eBNW
	 * @param[in] pnum p[eBN
	 * @param[out] vrts bV_
	 * @param[out] nrms _@
	 * @param[out] polys bV
	 * @param[in] filtering tB^OtO(fvXl0x01, ֊s0x02)
	 * @param[in] debug_output ʂ̉ʏo̗͂L
	 */
	virtual void CreateMesh(double *proj, double *modelview, int W, int H, vector<Vec3> &prts, int pnum, 
							vector<Vec3> &vrts, vector<Vec3> &nrms, vector< vector<int> > &mesh, 
							int filtering = 0, int debug_output = 0) = 0;

	/*!
	 * }bvObhz̃TCYύX
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] spacing fvX}bṽTvOԊu
	 */
	virtual void Resize(int W, int H, int spacing = -1) = 0;

	/*!
	 * _@vZ
	 * @param[in] vrts _W
	 * @param[in] nvrts _
	 * @param[in] tris Op`|S􉽏
	 * @param[in] ntris Op`|S
	 * @param[out] nrms @
	 */
	virtual void CalVertexNormals(const vector<Vec3> &vrts, unsigned int nvrts, const vector< vector<int> > &tris, unsigned int ntris, 
								  vector<Vec3> &nrms){}

public:
	// OpenGL`
	virtual void DrawSilhouetteEdge(void){}
	virtual void DrawSSVertex(Vec3 node_color, Vec3 edge_color){}
	virtual void DrawMeshGrid(int grid, const Vec3 colors[]){}
	virtual void DrawField(double minpos[2], double maxpos[2]){}

	virtual void DrawSSEdge(void){}

public:
	// fobOp
	virtual void OutputGridInfo(int grid){}
	virtual void OutputGridVertexInfo(int grid){}
};


//-----------------------------------------------------------------------------
// Screen Space Mesh - CPUł̎
// MARK:rxSSMeshCPU
//-----------------------------------------------------------------------------
class rxSSMeshCPU : public rxSSMesh
{
protected:
	// e[uCfbNXXV֐|C^
	typedef void (rxSSMeshCPU::*FuncTableIndex)(int&, int&, int[], rxSSGrid*);

protected:
	// oϐ
	vector<RXREAL> m_vSSDMap;		//!< fvX}bv

	vector<rxSSParticle> m_vSSPrts;		//!< KWnł̃p[eBN

	vector<rxSSEdge> m_vSSEdge;			//!< ֊sGbW
	vector<rxSSVertex> m_vSSEdgeVertex;	//!< GbW_
	vector<rxSSVertex> m_vSSVertex;		//!< XN[Xy[XbV_

	vector<rxSSGrid> m_vSSMGrid;		//!< bVpObh
	vector<int> m_vMeshGrid;			//!< fobOp:bVObh

	vector< vector<double> > m_vFilter;	//!< pbinomialW

	FuncTableIndex m_FuncTableIndex[25];//!< e[uCfbNXXV֐|C^

protected:
	//! ftHgRXgN^
	rxSSMeshCPU(){}	// wȂŃIuWFNg쐬̂h

public:
	//! RXgN^
	rxSSMeshCPU(double zmax, double h, double r, int n_filter, int n_iter);

	//! fXgN^
	virtual ~rxSSMeshCPU();

public:
	//
	// ANZX\bh
	//
	
	//! fvX}bv
	virtual RXREAL* GetDepthMap(void);

	//! _
	virtual rxSSVertex* GetSSVertex(void);

	//! bVpObh
	virtual rxSSGrid GetMeshGrid(int idx);
	virtual rxSSGrid GetMeshGrid(int i, int j){ return GetMeshGrid(i+j*m_iNgx); }

public:
	//
	// ֐
	//

	/*!
	 * XN[Xy[XbV
	 * @param[in] proj OpenGLes
	 * @param[in] modelview OpenGLfr[s
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] prts p[eBNW
	 * @param[in] pnum p[eBN
	 * @param[out] vrts bV_
	 * @param[out] polys bV
	 * @param[in] filtering tB^OtO(fvXl0x01, ֊s0x02)
	 * @param[in] debug_output ʂ̉ʏo̗͂L
	 */
	virtual void CreateMesh(double *proj, double *modelview, int W, int H, vector<Vec3> &prts, int pnum, 
					        vector<Vec3> &vrts, vector< vector<int> > &mesh, int filtering = 0, int debug_output = 0);

	/*!
	 * XN[Xy[XbV(@vZ܂)
	 * @param[in] proj OpenGLes
	 * @param[in] modelview OpenGLfr[s
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] prts p[eBNW
	 * @param[in] pnum p[eBN
	 * @param[out] vrts bV_
	 * @param[out] nrms _@
	 * @param[out] polys bV
	 * @param[in] filtering tB^OtO(fvXl0x01, ֊s0x02)
	 * @param[in] debug_output ʂ̉ʏo̗͂L
	 */
	virtual void CreateMesh(double *proj, double *modelview, int W, int H, vector<Vec3> &prts, int pnum, 
							vector<Vec3> &vrts, vector<Vec3> &nrms, vector< vector<int> > &mesh, 
							int filtering = 0, int debug_output = 0);

	/*!
	 * }bvObhz̃TCYύX
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] spacing fvX}bṽTvOԊu
	 */
	virtual void Resize(int W, int H, int spacing = -1);

	/*!
	 * _@vZ
	 * @param[in] vrts _W
	 * @param[in] nvrts _
	 * @param[in] tris Op`|S􉽏
	 * @param[in] ntris Op`|S
	 * @param[out] nrms @
	 */
	virtual void CalVertexNormals(const vector<Vec3> &vrts, unsigned int nvrts, const vector< vector<int> > &tris, unsigned int ntris, vector<Vec3> &nrms);

	/*!
	 * p[eBNWƔa𓧎eϊăfvX}bv𐶐
	 * @param[in] P es
	 * @param[in] MV fr[s
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] prts p[eBNW
	 * @param[in] pnum p[eBN
	 */
	void CalDepthMap(const rxMatrix4 &P, const rxMatrix4 &MV, int W, int H, vector<Vec3> &prts, int pnum);

	/*!
	 * fvX}bvɕ{
	 * @param[inout] dmap fvX}bv
	 * @param[in] nx,ny }bv𑜓x
	 * @param[in] n_filter tB^O
	 */
	void ApplyDepthFilter(vector<RXREAL> &dmap, int nx, int ny, int n_filter);

	/*!
	 * ֊sɕ{
	 * @param[inout] ssvrts XN[Wnł̒_
	 * @param[in] polys bV(\钸_)
	 * @param[in] n_iter tB^O
	 */
	void ApplySilhoutteSmoothing(vector<rxSSVertex> &ssvrts, const vector< vector<int> > &polys, int n_iter);

	/*!
	 * ֊sGbW̌ofront edge vertex̌vZ
	 * @param[in] nx,ny bVpObh𑜓x
	 * @param[in] dw,dh bVpObh
	 * @param[in] dgrid bVpfvX}bv
	 * @param[in] ssprts Screen Spaceł̃p[eBN
	 * @param[in] W,H XN[𑜓x
	 * @return oꂽ֊sGbW
	 */
	int DetectSilhouetteEdgeVertex(int nx, int ny, double dw, double dh, const vector<RXREAL> &dgrid, 
								   const vector<rxSSParticle> &ssprts, int W, int H);

	/*!
	 * ֊sGbW̌o
	 * @param[in] nx,ny bVpObh𑜓x
	 * @param[in] dw,dh bVpObh
	 * @param[in] dgrid bVpfvX}bv
	 * @return oꂽ֊sGbW
	 */
	int DetectSilhouetteEdge(int nx, int ny, double dw, double dh, const vector<RXREAL> &dgrid);


	/*!
	 * m[h_
	 * @param[in] nx,ny bVpObh𑜓x
	 * @param[in] dw,dh bVpObh
	 * @param[in] dgrid bVpfvX}bv
	 * @return ꂽm[h_
	 */
	int CalNodeVertex(int nx, int ny, double dw, double dh, const vector<RXREAL> &dgrid);

	/*!
	 * GbW_
	 * @param[in] nx,ny bVpObh𑜓x
	 * @param[in] dw,dh bVpObh
	 * @param[in] dgrid bVpfvX}bv
	 * @param[in] edges ֊sGbWXg
	 * @return ꂽGbW_
	 */
	int CalEdgeVertex(int nx, int ny, double dw, double dh, const vector<RXREAL> &dgrid, const vector<rxSSEdge> &edges);

	/*!
	 * Op`bV
	 * @param[in] nx,ny bVpObh𑜓x
	 * @param[in] dgrid bVObh
	 * @param[out] polys Op`bV
	 * @param[in] vstart _CfbNX̎n_
	 * @return ꂽOp`bV
	 */
	int CalMesh(int nx, int ny, vector<rxSSGrid> &grid, vector< vector<int> > &polys, int vstart = 0);

protected:
	// 񕪒Tɂ钸_ʒuvZ
	double binarySearchDepth(Vec3 v1, Vec3 v2, Vec3 &vr, double zmax);

	// fvXl̎Q
	RXREAL getDepthValue(Vec3 x)
	{
		return GetDepthInterpT(x[0], x[1], m_iNgx+1, m_iNgy+1, m_fDmx, m_fDmy, m_vSSDMap);
		//return GetDepthNearestT(x[0], x[1], m_iNgx+1, m_iNgy+1, m_fDmx, m_fDmy, m_vSSDMap);
	}

	// bVe[uCfbNX̍XV
	void updateTableIndexAny(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE0N4(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE1(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE2N4(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE3N23(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE3N4(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE4N2(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE4N3(int &table_index, int &vrot, int v[], rxSSGrid *g);
	void updateTableIndexE4N4(int &table_index, int &vrot, int v[], rxSSGrid *g);


public:
	// OpenGL`
	virtual void DrawSilhouetteEdge(void);
	virtual void DrawSSVertex(Vec3 node_color, Vec3 edge_color);
	virtual void DrawMeshGrid(int grid, const Vec3 colors[]);
	virtual void DrawField(double minpos[2], double maxpos[2]);

	virtual void DrawSSEdge(void);

public:
	// fobOp
	virtual void OutputGridInfo(int grid);
	virtual void OutputGridVertexInfo(int grid);

	int Mesh2Grid(int mesh_idx){ return m_vMeshGrid[mesh_idx]; }

};


//-----------------------------------------------------------------------------
// Screen Space Mesh - GPUł̎
// MARK:rxSSMeshGPU
//-----------------------------------------------------------------------------
class rxSSMeshGPU : public rxSSMesh
{
protected:
	// oϐ
	vector<float> m_vBinomials;			//!< pbinomialW(1D)

	vector<RXREAL> m_vSSDMap;			//!< fvX}bv
	vector<rxSSVertex> m_vSSVertex;		//!< XN[Xy[XbV_

	// CUDApϐ
	RXREAL *m_dSSPrtsCen;				//!< XN[Wnł̃p[eBNW
	RXREAL *m_dSSPrtsRad;				//!< XN[Wnł̃p[eBNa

	RXREAL *m_dSSDMap;					//!< fvX}bv
	rxVPacke m_dSSEdge;					//!< ֊sGbW
	rxVPackf m_dSSNodeVrts;				//!< m[h_
	rxVPackf m_dSSEdgeFrontVrts;		//!< GbW_(front)
	rxVPackf m_dSSEdgeBackVrts;			//!< GbW_(back)
	rxVPackf m_dSSEdgeBack2Vrts;		//!< GbW_(back-2)
	RXREAL *m_dSSVertex;				//!< S_(m[h_COʃGbW_CwʃGbW_CŔwʃGbW_)
	rxSSGridG *m_dSSMGrid;				//!< bVpObh
	unsigned int *m_dSSTriNum;					//!< Obh̃bV
	unsigned int *m_dSSTriNumScan;				//!< Obh̃bVScan
	//unsigned int *m_dSSTriArray;				//!< Op`bV

	RXREAL *m_hSSPrtsCen;				//!< XN[Wnł̃p[eBNW(foCXƂ̌p)
	RXREAL *m_hSSPrtsRad;				//!< XN[Wnł̃p[eBNa(foCXƂ̌p)

	rxSSEdgeG *m_hSSEdge;				//!< XN[Wnł̃p[eBNW(foCXƂ̌p)
	RXREAL *m_hSSVertex;				//!< XN[Wnł̃p[eBNW(foCXƂ̌p)

	rxSsmParams m_ssm_params;				//!< bVp[^

	// `VBO
	GLuint m_uVrtVBO;					//!< bV_(VBO)
	GLuint m_uTriVBO;					//!< bV|S(VBO)
	GLuint m_uNrmVBO;					//!< bV_@(VBO)


protected:
	//! ftHgRXgN^
	rxSSMeshGPU(){}	// wȂŃIuWFNg쐬̂h

public:
	//! RXgN^
	rxSSMeshGPU(double zmax, double h, double r, int n_filter, int n_iter);

	//! fXgN^
	virtual ~rxSSMeshGPU();

public:
	//
	// ANZX\bh
	//
	
	//! fvX}bv
	virtual RXREAL* GetDepthMap(void);

	//! _
	virtual rxSSVertex* GetSSVertex(void);

	//! bVpObh
	virtual rxSSGrid GetMeshGrid(int idx);
	virtual rxSSGrid GetMeshGrid(int i, int j){ return GetMeshGrid(i+j*m_iNgx); }

public:
	//
	// ֐
	//

	/*!
	 * XN[Xy[XbV
	 * @param[in] proj OpenGLes
	 * @param[in] modelview OpenGLfr[s
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] prts p[eBNW
	 * @param[in] pnum p[eBN
	 * @param[out] vrts bV_
	 * @param[out] polys bV
	 * @param[in] filtering tB^OtO(fvXl0x01, ֊s0x02)
	 * @param[in] debug_output ʂ̉ʏo̗͂L
	 */
	virtual void CreateMesh(double *proj, double *modelview, int W, int H, vector<Vec3> &prts, int pnum, 
							vector<Vec3> &vrts, vector< vector<int> > &mesh, 
							int filtering = 0, int debug_output = 0);

	/*!
	 * XN[Xy[XbV(@vZ܂)
	 * @param[in] proj OpenGLes
	 * @param[in] modelview OpenGLfr[s
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] prts p[eBNW
	 * @param[in] pnum p[eBN
	 * @param[out] vrts bV_
	 * @param[out] nrms _@
	 * @param[out] polys bV
	 * @param[in] filtering tB^OtO(fvXl0x01, ֊s0x02)
	 * @param[in] debug_output ʂ̉ʏo̗͂L
	 */
	virtual void CreateMesh(double *proj, double *modelview, int W, int H, vector<Vec3> &prts, int pnum, 
							vector<Vec3> &vrts, vector<Vec3> &nrms, vector< vector<int> > &mesh, 
							int filtering = 0, int debug_output = 0);

	void CreateMeshD(double *proj, double *modelview, int W, int H, RXREAL *dppos, int pnum, 
					 RXREAL* &dvrt, unsigned int* &dtri, int filtering = 0, int debug_output = 0);

	/*!
	 * }bvObhz̃TCYύX
	 * @param[in] W,H ʉ𑜓x
	 * @param[in] spacing fvX}bṽTvOԊu
	 */
	virtual void Resize(int W, int H, int spacing = -1);

	/*!
	 * _@vZ
	 * @param[in] vrts _W
	 * @param[in] nvrts _
	 * @param[in] tris Op`|S􉽏
	 * @param[in] ntris Op`|S
	 * @param[out] nrms @
	 */
	virtual void CalVertexNormals(const vector<Vec3> &vrts, unsigned int nvrts, const vector< vector<int> > &tris, unsigned int ntris, 
								  vector<Vec3> &nrms);

protected:
	//! GPŨp[^t@CXV
	void updateParams(int W, int H, const rxMatrix4 &P, const rxMatrix4 &PMV);

	void calVertexNormalD(RXREAL* dvrt, unsigned int nvrts, unsigned int* dtri, unsigned int ntris, RXREAL* &nrms);

public:
	// OpenGL`
	virtual void DrawSilhouetteEdge(void);
	virtual void DrawSSVertex(Vec3 node_color, Vec3 edge_color);
	virtual void DrawMeshGrid(int grid, const Vec3 colors[]);
	virtual void DrawField(double minpos[2], double maxpos[2]);

	virtual void DrawSSEdge(void);

public:
	// fobOp
	virtual void OutputGridInfo(int grid);
	virtual void OutputGridVertexInfo(int grid);

};







//-----------------------------------------------------------------------------
// ֐
//-----------------------------------------------------------------------------
/*!
 * Nv
 * @param[inout] x l
 * @param[in] a,b [a,b]ŃNv
 */
template<class T> 
inline void RX_CLAMP(T &x, const T &a, const T &b){ x = (x < a ? a : (x > b ? b : x)); }

inline void Swap(Vec4 &a, Vec4 &b)
{
	Vec4 tmp = a;
	a = b;
	b = tmp;
}


inline RXREAL GetDepthInterpT(double x, double y, int nx, int ny, double dx, double dy, const vector<RXREAL> &dmap)
{
	int ix0 = (int)(x/dx);
	int ix1 = ix0+1;
	double ddx = x/dx-ix0;

	RX_CLAMP(ix0, 0, nx-1);
	RX_CLAMP(ix1, 0, nx-1);

	int iy0 = (int)(y/dx);
	int iy1 = iy0+1;
	double ddy = y/dy-iy0;

	RX_CLAMP(iy0, 0, ny-1);
	RX_CLAMP(iy1, 0, ny-1);

	RXREAL d00 = dmap[iy0*nx+ix0];
	RXREAL d01 = dmap[iy0*nx+ix1];
	RXREAL d10 = dmap[iy1*nx+ix0];
	RXREAL d11 = dmap[iy1*nx+ix1];

	return (d00*(1.0-ddx)+d01*ddx)*(1.0-ddy)+(d10*(1.0-ddx)+d11*ddx)*ddy;
}


inline RXREAL GetDepthNearestT(double x, double y, int nx, int ny, double dx, double dy, const vector<RXREAL> &dmap)
{
	int ix = (int)(x/dx);
	RX_CLAMP(ix, 0, nx-1);

	int iy = (int)(y/dx);
	RX_CLAMP(iy, 0, ny-1);

	return dmap[iy*nx+ix];
}

// 񕪒T(1D)
inline double rtbis(double func(const double), const double x1, const double x2, const double xacc)
{
	const int JMAX = 40;
	double dx, f, fmid, xmid, rtb;

	f = func(x1);
	fmid = func(x2);
	if(f*fmid >= 0.0) return 0.0;

	rtb = f < 0.0 ? (dx = x2-x1, x1) : (dx = x1-x2, x2);
	for(int j = 0; j < JMAX; ++j){
		dx *= 0.5;
		xmid = rtb+dx;

		fmid = func(xmid);

		if(fmid <= 0.0){
			rtb = xmid;
		}

		if(fabs(dx) < xacc || fmid == 0.0){
			return rtb;
		}
	}
	
	return 0.0;
}


/*!
 * OpenGLϊsrxMatrix4ɕϊ
 *  column major  row major ɕϊrxMatrix4Ɋi[
 * @param[in] m OpenGLϊs(	glGetDoublev(GL_PROJECTION_MATRIX, m); ȂǂŎ擾)
 * @return ϊ̍s
 */
inline rxMatrix4 GetMatrixGL(double *m)
{
	return rxMatrix4(m[0], m[4], m[8],  m[12], 
					 m[1], m[5], m[9],  m[13], 
					 m[2], m[6], m[10], m[14], 
					 m[3], m[7], m[11], m[15]);
}


/*!
 * 2iɕϊ
 * @param[in] x ̐
 * @param[in] bit 2i
 * @return 2i
 */
inline string GetBitArray(int x, int bit)
{
    string s;
    s.resize(bit, '0');
    for(int i = 0; i < bit; ++i){
        s[bit-i-1] = ((x >> i) & 0x01) ? '1' : '0';
    }
    return s;
}

/*!
 * fvXl[0,1]ɕϊ
 * @param[in] depth fvXl
 * @param[in] w divisionl
 * @return [0,1]̒l
 */
inline double RX_DEPTH2COLORf(RXREAL depth)
{
	if(depth == RX_FEQ_INF){
		depth = 1.0;
	}
	else{
		depth *= 0.5;
	}
	RX_CLAMP(depth, (RXREAL)0.0, (RXREAL)1.0);
	return 1.0-depth;
}
inline unsigned char RX_DEPTH2COLOR(RXREAL d){ return (unsigned char)(RX_DEPTH2COLORf(d)*255); }


/*!
 * pbinominalWvZ
 * @param[in] b 
 * @return binominalW
 */
static vector< vector<double> > CalBinomials(int b)
{
	vector< vector<double> > bs;
	vector<double> f, tmp;
	f.resize(b+1);
	tmp.resize(b+1);

	bs.resize(b+1);

	double a = 1.0;

	for(int i = 0; i < b+1; ++i){
		f[i]   = (i == 0 ? 1 : 0);
		tmp[i] = (i == 0 ? 1 : 0);	
	}

	for(int k = 0; k < b+1; ++k){
		for(int i = 1; i < k+1; ++i){
			tmp[i] = f[i-1]+f[i];
		}

		for(int i = 1; i < k+1; ++i){
			f[i] = tmp[i];
		}

		bs[k].resize(k+1);
		for(int i = 0; i < k+1; ++i){
			bs[k][i] = f[i]*a;
		}

		a *= 0.5;
	}

	return bs;
}

/*!
 * pbinominalWvZ(1z, tB^ɗp镨̂)
 * @param[in] r tB^(b = 2*r+1)
 * @return binominalW
 */
static vector<float> CalBinomialsForFilter(int r)
{
	int b = 2*r+1;
	vector<float> bs;
	vector<float> f, tmp;
	f.resize(b+1);
	tmp.resize(b+1);

	bs.resize((r+1)*(r+1));

	float a = 1.0;

	for(int i = 0; i < b+1; ++i){
		f[i]   = (i == 0 ? 1.0f : 0.0f);
		tmp[i] = (i == 0 ? 1.0f : 0.0f);	
	}

	int c = 0;
	for(int k = 0; k < b+1; ++k){
		for(int i = 1; i < k+1; ++i){
			tmp[i] = f[i-1]+f[i];
		}

		for(int i = 1; i < k+1; ++i){
			f[i] = tmp[i];
		}

		if(!(k%2)){
			for(int i = 0; i < k+1; ++i){
				bs[c++] = f[i]*a;
			}
		}

		a *= 0.5;
	}

	return bs;
}

/*!
 * Ɖ~̌(2D)
 * @param[in] A,B ̗[_W
 * @param[in] C ~̒S
 * @param[in] r ~̔a
 * @param[out] P _W
 * @return _
 */
static int LineCircleIntersection(const Vec2 &A, const Vec2 &B, const Vec2 &C, const double &r, Vec2 P[2], double t[2])
{
	double rr = r*r;
	Vec2 AC = C-A;
	Vec2 BC = C-B;

	Vec2 v = B-A;
	double l = norm(v);
	v /= l;

	double td = dot(v, AC);
	Vec2 D = A+td*v;
	double dd = norm2(D-C);

	if(dd < rr){
		double dt = sqrt(rr-dd);

		double da = rr-norm2(AC);
		double db = rr-norm2(BC);

		int inter = 0;
		double t1 = td-dt;
		double t2 = td+dt;
		if(t1 >= 0 && t1 <= l){
			P[inter] = A+t1*v;
			t[inter] = t1;
			inter++;
		}
		if(t2 >= 0 && t2 <= l){
			P[inter] = A+t2*v;
			t[inter] = t2;
			inter++;
		}

		return inter;
	}
	else{
		return 0;
	}
}




#endif // #ifndef _RX_SSM_H_
